import codecs
import time
import webbrowser
from concurrent import futures
from datetime import datetime
from pathlib import Path

import click
import feedparser

try:
    from yaml import CLoader as Loader, dump, load
except ImportError:
    from yaml import Loader, dump, load


class Subscriptions:
    def __init__(self):
        fname = Path(Path.home() / "feed.yaml")
        if not fname.exists():
            fname.touch()
        self.f = fname

    def initialize(self):
        if self.f.exists():
            self.f.unlink()
        self.f.touch()

    def _fetch(self):
        f = codecs.open(self.f, mode="r", encoding="utf-8")
        return load(f, Loader=Loader)

    def _parse_title(self, url):
        d = feedparser.parse(url)
        title = d.feed.title
        return title

    def _update(self, data):
        f = codecs.open(self.f, mode="w", encoding="utf-8")
        dump(data, f, encoding="utf8", allow_unicode=True)

    def add(self, url, category):
        s = self._fetch()
        title = self._parse_title(url)
        if not s:
            # create a new category if no entries exists
            # then add a new subscription
            data = {category: [{"url": url, "title": title}]}
            self._update(data)
            return f"\nSubscribed: {title} to `{category}` category\n"
        else:
            subs = self._get_a_subscription(s)
            if subs:
                # check if subsriptions already exist
                if url in [sub["url"] for sub in subs]:
                    return f"\n`{url}` has already subscribed\n"
            if category not in s:
                s[category] = [{"url": url, "title": title}]
            else:
                s[category].append({"url": url, "title": title})
            self._update(s)
            return f"\nSubscribed: {title} to `{category}` category\n"

    def _get_categories(self, data):
        cats = [cat for cat in data.keys()]
        return cats

    def _get_a_subscription(self, data):
        subs = [sub for subscription in data.values() for sub in subscription]
        return subs

    def _get_all_subscriptions(self, data):
        subs = {sub[0]: sub[1] for sub in data.items()}
        return subs

    def lists(self, category, subscription):
        s = self._fetch()
        if not s:
            return False
        else:
            if category:
                return self._get_categories(s)
            elif subscription:
                return self._get_a_subscription(s)
            else:
                return self._get_all_subscriptions(s)

    def _show_category(self, data):
        cats = self._get_categories(data)
        while True:
            for i, cat in enumerate(cats, 1):
                click.echo(f"{i:^3} {cat}")
            click.echo()
            num = click.prompt(
                "Choose a category you want to view [1{}]".format(
                    "-" + str(len(cats)) if len(cats) > 1 else ""
                ),
                type=int,
            )
            try:
                subs = data[cats[num - 1]]
            except IndexError:
                click.echo("ERROR: Input number is out of range.\n")
            else:
                return cats[num - 1], [sub["url"] for sub in subs]
                break

    def _show_subscription(self, data):
        subs = self._get_a_subscription(data)
        while True:
            for i, sub in enumerate(subs, 1):
                click.echo("{:^3} {}".format(i, sub["title"]))
            click.echo()
            num = click.prompt(
                "Choose a subscription you want to view [1{}]".format(
                    "-" + str(len(subs)) if len(subs) > 1 else ""
                ),
                type=int,
            )
            try:
                sub = subs[num - 1]
            except IndexError:
                click.echo("ERROR: Input number is out of range.\n")
            else:
                return sub["title"], [sub["url"]]
                break

    def _show_all(self, data):
        subs = self._get_all_subscriptions(data)
        return "all", [s["url"] for sub in subs for s in subs[sub]]

    def show(self, category, subscription):
        s = self._fetch()
        if not s:
            return False
        else:
            if category:
                return self._show_category(s)
            elif subscription:
                return self._show_subscription(s)
            else:
                return self._show_all(s)

    def delete(self):
        s = self._fetch()
        if not s:
            return False
        else:
            cats = self._get_categories(s)
            while True:
                for i, cat in enumerate(cats, 1):
                    click.echo(f"{i:^3} {cat}")
                click.echo()
                num = click.prompt(
                    "Choose a category you want to delete [1{}]".format(
                        "-" + str(len(cats)) if len(cats) > 1 else ""
                    ),
                    type=int,
                )
                try:
                    del s[cats[num - 1]]
                except IndexError:
                    click.echo("ERROR: Input number is out of range.\n")
                else:
                    self._update(s)
                    return f"\nDeleted: `{cats[num - 1]}` category\n"
                    break

    def cancel(self):
        s = self._fetch()
        if not s:
            return False
        else:
            subs = self._get_a_subscription(s)
            while True:
                for i, sub in enumerate(subs, 1):
                    click.echo(f"{i:^3} {sub['title']}")
                click.echo()
                num = click.prompt(
                    "Choose a subscription you want to cancel [1{}]".format(
                        "-" + str(len(subs)) if len(subs) > 1 else ""
                    ),
                    type=int,
                )
                try:
                    sub = subs[num - 1]
                except IndexError:
                    click.echo("ERROR: Input number is out of range.\n")
                else:
                    # delete a subscription
                    for subs in s.values():
                        if sub in subs:
                            subs.remove(sub)
                    # delete category if no subscriptions exist
                    data = {cat: s[cat] for cat in s if s[cat]}
                    self._update(data)
                    return f"\nCanceled: {sub['title']}\n"
                    break


def when_create(updated):
    now = datetime.utcnow()
    between = now - updated
    day = between.days
    if day:
        if day >= 365:
            return f"{day // 365}y ago"
        if day == 1:
            return "yesterday"
        return f"{day}d ago"
    else:
        minute, second = divmod(between.seconds, 60)
        hour, minute_ = divmod(minute, 60)
        if hour:
            if hour > 1:
                return f"{hour}h ago"
            else:
                return f"{hour}h ago"
        elif minute_:
            if minute_ > 1:
                return f"{minute_}m ago"
            else:
                return f"{minute_}m ago"
        else:
            if second > 1:
                return f"{second}s ago"
            else:
                return f"{second}s ago"


def _parse_feed(url):
    d = feedparser.parse(url)
    return [
        (
            entry.title,
            entry.link,
            datetime.strptime(
                time.strftime("%Y/%m/%d-%H:%M:%S", entry.updated_parsed),
                "%Y/%m/%d-%H:%M:%S",
            ),
        )
        for entry in d.entries
    ]


def _list_articles(subs):
    e = futures.ThreadPoolExecutor()
    fs = [e.submit(_parse_feed, url) for url in subs[1]]
    f = [f.result() for f in futures.as_completed(fs)]
    articles = sorted(
        [f for fs in f for f in fs], key=lambda updated: updated[2], reverse=True
    )

    # yield n-sized chunks (in this case, n is 30)
    articles = [articles[i : i + 30] for i in range(0, len(articles), 30)]
    _i = 0

    click.clear()
    if subs[0] == "all":
        click.echo("\nAll articles\n")
    else:
        click.echo(f"\nArticles::{subs[0]}\n")

    def _recursive(_i, articles):
        for i, article in enumerate(articles[_i], 1):
            click.echo(f"{i:^3} {when_create(article[2]):^10} {article[0]}")
        click.echo()

        while True:
            try:
                num = click.prompt(
                    "Input [1{}] to open browser {}or [Ctrl+C] to exit".format(
                        "-" + str(len(articles[_i])) if len(articles[_i]) > 1 else "",
                        "or [n] to read more " if len(articles[_i]) == 30 else "",
                    )
                )
                if num == "n":
                    _i = _i + 1
                    try:
                        articles[_i]
                    except IndexError:
                        exit()
                    else:
                        # read next chunk
                        click.clear()
                        click.echo()
                        _recursive(_i, articles)
                else:
                    try:
                        url = articles[_i][int(num) - 1][1]
                    except IndexError:
                        click.echo("ERROR: Input number is out of range.\n")
                    except ValueError:
                        click.echo(f"ERROR: {num} is not a valid integer.\n")
                    else:
                        webbrowser.open(url)
            except KeyboardInterrupt:
                exit()

    _recursive(_i, articles)


def validate_rss_url(ctx, param, url):
    d = feedparser.parse(url)
    if not d["entries"]:
        raise click.BadParameter(f"`{url}` is non-well-formed feed\n")
    else:
        return url


def print_help(command):
    with click.Context(command) as ctx:
        click.echo(command.get_help(ctx))


@click.group()
def teader():
    """A Feed reader from terminal"""
    pass


@teader.command()
def init():
    """initialize subscriptions"""
    Subscriptions().initialize()
    click.echo("\nInitialize database\n")


@teader.command()
@click.argument("url", metavar="<RSS URL>", callback=validate_rss_url)
@click.option(
    "-c",
    "--category",
    type=str,
    metavar="<CATEGORY>",
    default="uncategorized",
    help="Add a subscription to <CATEGORY>",
)
def add(url, category):
    """Add a subscription"""
    click.echo(Subscriptions().add(url, category))


@teader.command()
@click.option("-c", "--category", is_flag=True, help="Show articles by category")
@click.option(
    "-s", "--subscription", is_flag=True, help="Show articles by subscription"
)
def show(category, subscription):
    """Show articles"""
    subs = Subscriptions().show(category, subscription)
    if not subs:
        click.echo(f"\nNo subscriptions\n")
    else:
        _list_articles(subs)


@teader.command()
@click.option(
    "-c", "--category", is_flag=True, default=False, help="List all categories"
)
@click.option(
    "-s",
    "--subscription",
    is_flag=True,
    default=False,
    help="List subscriptions by category",
)
def list(category, subscription):
    """Show a list of subscriptions"""
    subs = Subscriptions().lists(category, subscription)
    if not subs:
        click.echo(f"\nNo subscriptions\n")
    else:
        if category:
            click.echo(f"\nAll categories:\n")
            for sub in subs:
                click.echo(f"- {sub}")
        elif subscription:
            click.echo(f"\nAll subscriptions:\n")
            for sub in subs:
                click.echo(f"- {sub['title']}")
        else:
            click.echo(f"\nAll subscriptions:\n")
            for cat in subs.keys():
                click.echo(f"[{cat}]")
                for sub in subs[cat]:
                    click.echo(f"- {sub['title']}")


@teader.command()
@click.option("-c", "--category", is_flag=True, help="Delete a category")
@click.option("-s", "--subscription", is_flag=True, help="Cancel a subscription")
def delete(category, subscription):
    """Delete a category or Cancel a subscription"""
    if category:
        cats = Subscriptions().delete()
        if not cats:
            click.echo(f"\nNo subscriptions\n")
        else:
            click.echo(cats)
    elif subscription:
        sub = Subscriptions().cancel()
        if not sub:
            click.echo(f"\nNo subscriptions\n")
        else:
            click.echo(sub)
    else:
        print_help(delete)
